"""Handles username/password authentication and two-step authentication"""

import logging
import sys
import time
from typing import Callable, Dict, Optional, Tuple

import click

from icloudpd.mfa_provider import MFAProvider
from icloudpd.status import Status, StatusExchange
from pyicloud_ipd.base import PyiCloudService
from pyicloud_ipd.file_match import FileMatchPolicy
from pyicloud_ipd.raw_policy import RawTreatmentPolicy


class TwoStepAuthRequiredError(Exception):
    """
    Raised when 2SA is required. base.py catches this exception
    and sends an email notification.
    """


def authenticator(
    logger: logging.Logger,
    domain: str,
    filename_cleaner: Callable[[str], str],
    lp_filename_generator: Callable[[str], str],
    raw_policy: RawTreatmentPolicy,
    file_match_policy: FileMatchPolicy,
    password_providers: Dict[
        str, Tuple[Callable[[str], Optional[str]], Callable[[str, str], None]]
    ],
    mfa_provider: MFAProvider,
    status_exchange: StatusExchange,
) -> Callable[[str, Optional[str], bool, Optional[str]], PyiCloudService]:
    """Wraping authentication with domain context"""

    def authenticate_(
        username: str,
        cookie_directory: Optional[str] = None,
        raise_error_on_2sa: bool = False,
        client_id: Optional[str] = None,
    ) -> PyiCloudService:
        """Authenticate with iCloud username and password"""
        logger.debug("Authenticating...")
        icloud: Optional[PyiCloudService] = None
        _valid_password: Optional[str] = None
        for _, _pair in password_providers.items():
            _reader, _ = _pair
            _password = _reader(username)
            if _password:
                icloud = PyiCloudService(
                    filename_cleaner,
                    lp_filename_generator,
                    domain,
                    raw_policy,
                    file_match_policy,
                    username,
                    _password,
                    cookie_directory=cookie_directory,
                    client_id=client_id,
                )
                _valid_password = _password
                break

        if not icloud:
            raise NotImplementedError("None of providers gave password")

        if _valid_password:
            # save valid password to all providers
            for _, _pair in password_providers.items():
                _, _writer = _pair
                _writer(username, _valid_password)

        if icloud.requires_2fa:
            if raise_error_on_2sa:
                raise TwoStepAuthRequiredError("Two-factor authentication is required")
            logger.info("Two-factor authentication is required (2fa)")
            if mfa_provider == MFAProvider.WEBUI:
                request_2fa_web(icloud, logger, status_exchange)
            else:
                request_2fa(icloud, logger)

        elif icloud.requires_2sa:
            if raise_error_on_2sa:
                raise TwoStepAuthRequiredError("Two-step authentication is required")
            logger.info("Two-step authentication is required (2sa)")
            request_2sa(icloud, logger)

        return icloud

    return authenticate_


def request_2sa(icloud: PyiCloudService, logger: logging.Logger) -> None:
    """Request two-step authentication. Prompts for SMS or device"""
    devices = icloud.trusted_devices
    devices_count = len(devices)
    device_index: int = 0
    if devices_count > 0:
        for i, device in enumerate(devices):
            number = device["phoneNumber"]
            alt_name = f"SMS to {number}"
            name = device.get("deviceName", alt_name)
            click.echo(f"  {i}: {name}")

        device_index = click.prompt(
            "Please choose an option:", default="0", type=click.IntRange(0, devices_count - 1)
        )

    device = devices[device_index]
    if not icloud.send_verification_code(device):
        logger.error("Failed to send two-step authentication code")
        sys.exit(1)

    code = click.prompt("Please enter two-step authentication code")
    if not icloud.validate_verification_code(device, code):
        logger.error("Failed to verify two-step authentication code")
        sys.exit(1)
    logger.info(
        "Great, you're all set up. The script can now be run without "
        "user interaction until 2SA expires.\n"
        "You can set up email notifications for when "
        "the two-step authentication expires.\n"
        "(Use --help to view information about SMTP options.)"
    )


def request_2fa(icloud: PyiCloudService, logger: logging.Logger) -> None:
    """Request two-factor authentication."""
    devices = icloud.get_trusted_phone_numbers()
    devices_count = len(devices)
    device_index_alphabet = "abcdefghijklmnopqrstuvwxyz"
    if devices_count > 0:
        if devices_count > len(device_index_alphabet):
            logger.error("Too many trusted devices for authentication")
            sys.exit(1)

        for i, device in enumerate(devices):
            click.echo(f"  {device_index_alphabet[i]}: {device.obfuscated_number}")

        index_str = f"..{device_index_alphabet[devices_count - 1]}" if devices_count > 1 else ""
        index_or_code: str = ""
        while True:
            index_or_code = (
                click.prompt(
                    f"Please enter two-factor authentication code or device index ({device_index_alphabet[0]}{index_str}) to send SMS with a code",
                )
                .strip()
                .lower()
            )

            if index_or_code == "":
                click.echo("Empty string. Try again")
                continue

            if len(index_or_code) == 1:
                if index_or_code in device_index_alphabet:
                    if device_index_alphabet.index(index_or_code) > devices_count - 1:
                        click.echo(
                            f"Invalid index, should be ({device_index_alphabet[0]}{index_str}). Try again",
                        )
                        continue
                    else:
                        break
                else:
                    click.echo(
                        f"Invalid index, should be ({device_index_alphabet[0]}{index_str}). Try again",
                    )
                    continue

            if len(index_or_code) == 6:
                if index_or_code.isdigit():
                    break
                else:
                    click.echo("Invalid code, should be six digits. Try again")
                    continue

            click.echo(
                f"Should be index ({device_index_alphabet[0]}{index_str}) or six-digit code. Try again",
            )

        if index_or_code in device_index_alphabet:
            # need to send code
            device_index = device_index_alphabet.index(index_or_code)
            device = devices[device_index]
            if not icloud.send_2fa_code_sms(device.id):
                logger.error("Failed to send two-factor authentication code")
                sys.exit(1)
            while True:
                code: str = click.prompt(
                    "Please enter two-factor authentication code that you received over SMS",
                ).strip()
                if len(code) == 6 and code.isdigit():
                    break
                click.echo("Invalid code, should be six digits. Try again")

            if not icloud.validate_2fa_code_sms(device.id, code):
                logger.error("Failed to verify two-factor authentication code")
                sys.exit(1)
        else:
            if not icloud.validate_2fa_code(index_or_code):
                logger.error("Failed to verify two-factor authentication code")
                sys.exit(1)
    else:
        while True:
            code = click.prompt(
                "Please enter two-factor authentication code",
            ).strip()
            if len(code) == 6 and code.isdigit():
                break
            click.echo("Invalid code, should be six digits. Try again")
        if not icloud.validate_2fa_code(code):
            logger.error("Failed to verify two-factor authentication code")
            sys.exit(1)
    logger.info(
        "Great, you're all set up. The script can now be run without "
        "user interaction until 2FA expires.\n"
        "You can set up email notifications for when "
        "the two-factor authentication expires.\n"
        "(Use --help to view information about SMTP options.)"
    )


def request_2fa_web(
    icloud: PyiCloudService, logger: logging.Logger, status_exchange: StatusExchange
) -> None:
    """Request two-factor authentication through Webui."""
    if not status_exchange.replace_status(Status.NO_INPUT_NEEDED, Status.NEED_MFA):
        logger.error("Expected NO_INPUT_NEEDED, but got something else")
        return

    # wait for input
    while True:
        status = status_exchange.get_status()
        if status == Status.NEED_MFA:
            time.sleep(1)
        else:
            break

    if status_exchange.replace_status(Status.SUPPLIED_MFA, Status.CHECKING_MFA):
        code = status_exchange.get_payload()
        if not code:
            logger.error("Internal error: did not get code for SUPPLIED_MFA status")
            status_exchange.replace_status(
                Status.CHECKING_MFA, Status.NO_INPUT_NEEDED
            )  # TODO Error
            return

        if not icloud.validate_2fa_code(code):
            logger.error("Failed to verify two-factor authentication code")
            status_exchange.replace_status(
                Status.CHECKING_MFA, Status.NO_INPUT_NEEDED
            )  # TODO Error
            return
        status_exchange.replace_status(Status.CHECKING_MFA, Status.NO_INPUT_NEEDED)  # done

        logger.info(
            "Great, you're all set up. The script can now be run without "
            "user interaction until 2FA expires.\n"
            "You can set up email notifications for when "
            "the two-factor authentication expires.\n"
            "(Use --help to view information about SMTP options.)"
        )
    else:
        logger.error("Failed to change status")



"""Atomic counter"""

from multiprocessing import Lock, RawValue


class Counter:
    def __init__(self, value: int = 0):
        self.initial_value = value
        self.val = RawValue("i", value)
        self.lock = Lock()

    def increment(self) -> None:
        with self.lock:
            self.val.value += 1

    def reset(self) -> None:
        with self.lock:
            self.val = RawValue("i", self.initial_value)

    def value(self) -> int:
        with self.lock:
            return int(self.val.value)



import copy
import os
import sys
import json
import logging
import base64
import re

from datetime import datetime
from typing import Any, Callable, Dict, Generator, Optional, Sequence, Tuple, TypeVar, Union, cast
import typing

from requests import Response
from foundation import wrap_param_in_exception, bytes_decode
from foundation.core import compose, identity
from pyicloud_ipd.asset_version import AssetVersion
from pyicloud_ipd.exceptions import PyiCloudServiceNotActivatedException
from pyicloud_ipd.exceptions import PyiCloudAPIResponseException

import pytz

from urllib.parse import urlencode

from pyicloud_ipd.file_match import FileMatchPolicy
from pyicloud_ipd.item_type import AssetItemType
from pyicloud_ipd.raw_policy import RawTreatmentPolicy
from pyicloud_ipd.session import PyiCloudSession
from pyicloud_ipd.utils import add_suffix_to_filename
from pyicloud_ipd.version_size import AssetVersionSize, LivePhotoVersionSize, VersionSize

logger = logging.getLogger(__name__)


class PhotoLibrary(object):
    """Represents a library in the user's photos.

    This provides access to all the albums as well as the photos.
    """
    SMART_FOLDERS = {
        "All Photos": {
            "obj_type": "CPLAssetByAssetDateWithoutHiddenOrDeleted",
            "list_type": "CPLAssetAndMasterByAssetDateWithoutHiddenOrDeleted",
            "direction": "ASCENDING",
            "query_filter": None
        },
        "Time-lapse": {
            "obj_type": "CPLAssetInSmartAlbumByAssetDate:Timelapse",
            "list_type": "CPLAssetAndMasterInSmartAlbumByAssetDate",
            "direction": "ASCENDING",
            "query_filter": [{
                "fieldName": "smartAlbum",
                "comparator": "EQUALS",
                "fieldValue": {
                    "type": "STRING",
                    "value": "TIMELAPSE"
                }
            }]
        },
        "Videos": {
            "obj_type": "CPLAssetInSmartAlbumByAssetDate:Video",
            "list_type": "CPLAssetAndMasterInSmartAlbumByAssetDate",
            "direction": "ASCENDING",
            "query_filter": [{
                "fieldName": "smartAlbum",
                "comparator": "EQUALS",
                "fieldValue": {
                    "type": "STRING",
                    "value": "VIDEO"
                }
            }]
        },
        "Slo-mo": {
            "obj_type": "CPLAssetInSmartAlbumByAssetDate:Slomo",
            "list_type": "CPLAssetAndMasterInSmartAlbumByAssetDate",
            "direction": "ASCENDING",
            "query_filter": [{
                "fieldName": "smartAlbum",
                "comparator": "EQUALS",
                "fieldValue": {
                    "type": "STRING",
                    "value": "SLOMO"
                }
            }]
        },
        "Bursts": {
            "obj_type": "CPLAssetBurstStackAssetByAssetDate",
            "list_type": "CPLBurstStackAssetAndMasterByAssetDate",
            "direction": "ASCENDING",
            "query_filter": None
        },
        "Favorites": {
            "obj_type": "CPLAssetInSmartAlbumByAssetDate:Favorite",
            "list_type": "CPLAssetAndMasterInSmartAlbumByAssetDate",
            "direction": "ASCENDING",
            "query_filter": [{
                "fieldName": "smartAlbum",
                "comparator": "EQUALS",
                "fieldValue": {
                    "type": "STRING",
                    "value": "FAVORITE"
                }
            }]
        },
        "Panoramas": {
            "obj_type": "CPLAssetInSmartAlbumByAssetDate:Panorama",
            "list_type": "CPLAssetAndMasterInSmartAlbumByAssetDate",
            "direction": "ASCENDING",
            "query_filter": [{
                "fieldName": "smartAlbum",
                "comparator": "EQUALS",
                "fieldValue": {
                    "type": "STRING",
                    "value": "PANORAMA"
                }
            }]
        },
        "Screenshots": {
            "obj_type": "CPLAssetInSmartAlbumByAssetDate:Screenshot",
            "list_type": "CPLAssetAndMasterInSmartAlbumByAssetDate",
            "direction": "ASCENDING",
            "query_filter": [{
                "fieldName": "smartAlbum",
                "comparator": "EQUALS",
                "fieldValue": {
                    "type": "STRING",
                    "value": "SCREENSHOT"
                }
            }]
        },
        "Live": {
            "obj_type": "CPLAssetInSmartAlbumByAssetDate:Live",
            "list_type": "CPLAssetAndMasterInSmartAlbumByAssetDate",
            "direction": "ASCENDING",
            "query_filter": [{
                "fieldName": "smartAlbum",
                "comparator": "EQUALS",
                "fieldValue": {
                    "type": "STRING",
                    "value": "LIVE"
                }
            }]
        },
        "Recently Deleted": {
            "obj_type": "CPLAssetDeletedByExpungedDate",
            "list_type": "CPLAssetAndMasterDeletedByExpungedDate",
            "direction": "ASCENDING",
            "query_filter": None
        },
        "Hidden": {
            "obj_type": "CPLAssetHiddenByAssetDate",
            "list_type": "CPLAssetAndMasterHiddenByAssetDate",
            "direction": "ASCENDING",
            "query_filter": None
        },
    }

    def __init__(self, service: "PhotosService", zone_id: Dict[str, Any]):
        self.service = service
        self.zone_id = zone_id

        self._albums: Optional[Dict[str, PhotoAlbum]] = None

        url = ('%s/records/query?%s' %
               (self.service._service_endpoint, urlencode(self.service.params)))
        json_data = json.dumps({
            "query": {"recordType":"CheckIndexingState"},
            "zoneID": self.zone_id,
        })

        request = self.service.session.post(
            url,
            data=json_data,
            headers={'Content-type': 'text/plain'}
        )
        response = request.json()
        indexing_state = response['records'][0]['fields']['state']['value']
        if indexing_state != 'FINISHED':
            raise PyiCloudServiceNotActivatedException(
                ('iCloud Photo Library not finished indexing.  Please try '
                 'again in a few minutes'), None)

    @property
    def albums(self) -> Dict[str, "PhotoAlbum"]:
        if not self._albums:
            self._albums = {
                name: PhotoAlbum(self.service, name, zone_id=self.zone_id, **props) # type: ignore[arg-type] # dynamically builing params 
                for (name, props) in self.SMART_FOLDERS.items()
            }

            for folder in self._fetch_folders():
                # FIXME: Handle subfolders
                if folder['recordName'] in ('----Root-Folder----',
                    '----Project-Root-Folder----') or \
                    (folder['fields'].get('isDeleted') and
                     folder['fields']['isDeleted']['value']):
                    continue

                folder_id = folder['recordName']
                folder_obj_type = \
                    "CPLContainerRelationNotDeletedByAssetDate:%s" % folder_id
                folder_name = base64.b64decode(
                    folder['fields']['albumNameEnc']['value']).decode('utf-8')
                query_filter = [{
                    "fieldName": "parentId",
                    "comparator": "EQUALS",
                    "fieldValue": {
                        "type": "STRING",
                        "value": folder_id
                    }
                }]

                album = PhotoAlbum(self.service, folder_name,
                                   'CPLContainerRelationLiveByAssetDate',
                                   folder_obj_type, 'ASCENDING', query_filter,
                                   zone_id=self.zone_id)
                self._albums[folder_name] = album

        return self._albums

    def _fetch_folders(self) -> Sequence[Dict[str, Any]]:
        url = ('%s/records/query?%s' %
               (self.service._service_endpoint, urlencode(self.service.params)))
        json_data = json.dumps({
            "query": {"recordType":"CPLAlbumByPositionLive"},
            "zoneID": self.zone_id,
        })

        request = self.service.session.post(
            url,
            data=json_data,
            headers={'Content-type': 'text/plain'}
        )
        response = request.json()

        return typing.cast(Sequence[Dict[str, Any]], response['records'])

    @property
    def all(self) -> "PhotoAlbum":
        return self.albums['All Photos']


class PhotosService(PhotoLibrary):
    """The 'Photos' iCloud service.

    This also acts as a way to access the user's primary library.
    """
    def __init__(
            self, 
            service_root: str, 
            session: PyiCloudSession, 
            params: Dict[str, Any], 
            filename_cleaner:Callable[[str], str], 
            lp_filename_generator: Callable[[str], str], 
            raw_policy: RawTreatmentPolicy,
            file_match_policy: FileMatchPolicy):
        self.session = session
        self.params = dict(params)
        self._service_root = service_root
        self._service_endpoint = \
            ('%s/database/1/com.apple.photos.cloud/production/private'
             % self._service_root)

        self._libraries: Optional[Dict[str, PhotoLibrary]] = None

        self.filename_cleaner = filename_cleaner
        self.lp_filename_generator = lp_filename_generator
        self.raw_policy = raw_policy
        self.file_match_policy = file_match_policy

        self.params.update({
            'remapEnums': True,
            'getCurrentSyncToken': True
        })

        # TODO: Does syncToken ever change?
        # self.params.update({
        #     'syncToken': response['syncToken'],
        #     'clientInstanceId': self.params.pop('clientId')
        # })

        # self._photo_assets = {}

        super(PhotosService, self).__init__(
            service=self, zone_id={u'zoneName': u'PrimarySync'})

    @property
    def libraries(self) -> Dict[str, PhotoLibrary]:
        if not self._libraries:
            try:
                url = ('%s/zones/list' %
                    (self._service_endpoint, ))
                request = self.session.post(
                    url,
                    data='{}',
                    headers={'Content-type': 'text/plain'}
                )
                response = request.json()
                zones = response['zones'] 
            except Exception as e:
                    logger.error("library exception: %s" % str(e))

            libraries = {}
            for zone in zones:
                if not zone.get('deleted'):
                    zone_name = zone['zoneID']['zoneName']
                    libraries[zone_name] = PhotoLibrary(
                        self, zone_id=zone['zoneID'])
                        # obj_type='CPLAssetByAssetDateWithoutHiddenOrDeleted',
                        # list_type="CPLAssetAndMasterByAssetDateWithoutHiddenOrDeleted",
                        # direction="ASCENDING", query_filter=None,
                        # zone_id=zone['zoneID'])

            self._libraries = libraries

        return self._libraries


class PhotoAlbum(object):

    def __init__(self, service:PhotosService, name: str, list_type: str, obj_type: str, direction: str,
                 query_filter:Optional[Sequence[Dict[str, Any]]]=None, page_size:int=100, zone_id:Optional[Dict[str, Any]]=None):
        self.name = name
        self.service = service
        self.list_type = list_type
        self.obj_type = obj_type
        self.direction = direction
        self.query_filter = query_filter
        self.page_size = page_size
        self.exception_handler: Optional[Callable[[Exception, int], None]] = None

        self._len: Optional[int] = None

        if zone_id:
            self._zone_id: Dict[str, Any] = zone_id
        else:
            self._zone_id = {u'zoneName': u'PrimarySync'}

    @property
    def title(self) -> str:
        return self.name

    def __iter__(self) -> Generator["PhotoAsset", Any, None]:
        return self.photos

    def __len__(self) -> int:
        if self._len is None:
            url = ('%s/internal/records/query/batch?%s' %
                   (self.service._service_endpoint,
                    urlencode(self.service.params)))
            request = self.service.session.post(
                url,
                data=json.dumps(self._count_query_gen(self.obj_type)),
                headers={'Content-type': 'text/plain'}
            )
            response = request.json()

            self._len = (response["batch"][0]["records"][0]["fields"]
                         ["itemCount"]["value"])

        return self._len

    # Perform the request in a separate method so that we
    # can mock it to test session errors.
    def photos_request(self, offset: int) -> Response:
        url = ('%s/records/query?' % self.service._service_endpoint) + \
            urlencode(self.service.params)
        return self.service.session.post(
            url,
            data=json.dumps(self._list_query_gen(
                offset, self.list_type, self.direction,
                self.query_filter)),
            headers={'Content-type': 'text/plain'}
        )


    @property
    def photos(self) -> Generator["PhotoAsset", Any, None]:
        if self.direction == "DESCENDING":
            offset = len(self) - 1
        else:
            offset = 0

        exception_retries = 0

        while(True):
            try:
                request = self.photos_request(offset)
            except PyiCloudAPIResponseException as ex:
                if self.exception_handler:
                    exception_retries += 1
                    self.exception_handler(ex, exception_retries)
                    if exception_retries > 5:
                        raise
                    continue
                else:
                    raise

            exception_retries = 0

#            url = ('%s/records/query?' % self.service._service_endpoint) + \
#                urlencode(self.service.params)
#            request = self.service.session.post(
#                url,
#                data=json.dumps(self._list_query_gen(
#                    offset, self.list_type, self.direction,
#                    self.query_filter)),
#                headers={'Content-type': 'text/plain'}
#            )

            response = request.json()

            asset_records = {}
            master_records = []
            for rec in response['records']:
                if rec['recordType'] == "CPLAsset":
                    master_id = \
                        rec['fields']['masterRef']['value']['recordName']
                    asset_records[master_id] = rec
                elif rec['recordType'] == "CPLMaster":
                    master_records.append(rec)

            master_records_len = len(master_records)
            if master_records_len:
                if self.direction == "DESCENDING":
                    offset = offset - master_records_len
                else:
                    offset = offset + master_records_len

                for master_record in master_records:
                    record_name = master_record['recordName']
                    yield PhotoAsset(self.service, master_record,
                                     asset_records[record_name])
            else:
                break

    def _count_query_gen(self, obj_type: str) -> Dict[str, Any]:
        query = {
            u'batch': [{
                u'resultsLimit': 1,
                u'query': {
                    u'filterBy': {
                        u'fieldName': u'indexCountID',
                        u'fieldValue': {
                            u'type': u'STRING_LIST',
                            u'value': [
                                obj_type
                            ]
                        },
                        u'comparator': u'IN'
                    },
                    u'recordType': u'HyperionIndexCountLookup'
                },
                u'zoneWide': True,
                u'zoneID': self._zone_id
            }]
        }

        return query

    def _list_query_gen(self, offset: int, list_type: str, direction: str, query_filter:Optional[Sequence[Dict[str,None]]]=None) -> Dict[str, Any]:
        query: Dict[str, Any] = {
            u'query': {
                u'filterBy': [
                    {u'fieldName': u'startRank', u'fieldValue':
                        {u'type': u'INT64', u'value': offset},
                        u'comparator': u'EQUALS'},
                    {u'fieldName': u'direction', u'fieldValue':
                        {u'type': u'STRING', u'value': direction},
                        u'comparator': u'EQUALS'}
                ],
                u'recordType': list_type
            },
            u'resultsLimit': self.page_size * 2,
            u'desiredKeys': [
                u'resJPEGFullWidth', u'resJPEGFullHeight',
                u'resJPEGFullFileType', u'resJPEGFullFingerprint',
                u'resJPEGFullRes', u'resJPEGLargeWidth',
                u'resJPEGLargeHeight', u'resJPEGLargeFileType',
                u'resJPEGLargeFingerprint', u'resJPEGLargeRes',
                u'resJPEGMedWidth', u'resJPEGMedHeight',
                u'resJPEGMedFileType', u'resJPEGMedFingerprint',
                u'resJPEGMedRes', u'resJPEGThumbWidth',
                u'resJPEGThumbHeight', u'resJPEGThumbFileType',
                u'resJPEGThumbFingerprint', u'resJPEGThumbRes',
                u'resVidFullWidth', u'resVidFullHeight',
                u'resVidFullFileType', u'resVidFullFingerprint',
                u'resVidFullRes', u'resVidMedWidth', u'resVidMedHeight',
                u'resVidMedFileType', u'resVidMedFingerprint',
                u'resVidMedRes', u'resVidSmallWidth', u'resVidSmallHeight',
                u'resVidSmallFileType', u'resVidSmallFingerprint',
                u'resVidSmallRes', u'resSidecarWidth', u'resSidecarHeight',
                u'resSidecarFileType', u'resSidecarFingerprint',
                u'resSidecarRes', u'itemType', u'dataClassType',
                u'filenameEnc', u'originalOrientation', u'resOriginalWidth',
                u'resOriginalHeight', u'resOriginalFileType',
                u'resOriginalFingerprint', u'resOriginalRes',
                u'resOriginalAltWidth', u'resOriginalAltHeight',
                u'resOriginalAltFileType', u'resOriginalAltFingerprint',
                u'resOriginalAltRes', u'resOriginalVidComplWidth',
                u'resOriginalVidComplHeight', u'resOriginalVidComplFileType',
                u'resOriginalVidComplFingerprint', u'resOriginalVidComplRes',
                u'isDeleted', u'isExpunged', u'dateExpunged', u'remappedRef',
                u'recordName', u'recordType', u'recordChangeTag',
                u'masterRef', u'adjustmentRenderType', u'assetDate',
                u'addedDate', u'isFavorite', u'isHidden', u'orientation',
                u'duration', u'assetSubtype', u'assetSubtypeV2',
                u'assetHDRType', u'burstFlags', u'burstFlagsExt', u'burstId',
                u'captionEnc', u'locationEnc', u'locationV2Enc',
                u'locationLatitude', u'locationLongitude', u'adjustmentType',
                u'timeZoneOffset', u'vidComplDurValue', u'vidComplDurScale',
                u'vidComplDispValue', u'vidComplDispScale',
                u'keywordsEnc',u'extendedDescEnc',u'adjustedMediaMetaDataEnc',u'adjustmentSimpleDataEnc',
                u'vidComplVisibilityState', u'customRenderedValue',
                u'containerId', u'itemId', u'position', u'isKeyAsset'
            ],
            u'zoneID': self._zone_id
        }

        if query_filter:
            query['query']['filterBy'].extend(query_filter)

        return query

    def __unicode__(self) -> str:
        return self.title

    def __str__(self) -> str:
        as_unicode = self.__unicode__()
        if sys.version_info[0] >= 3:
            return as_unicode
        else:
            return as_unicode.encode('ascii', 'ignore')

    def __repr__(self) -> str:
        return "<%s: '%s'>" % (
            type(self).__name__,
            self
        )


class PhotoAsset(object):
    def __init__(self, service:PhotosService, master_record: Dict[str, Any], asset_record: Dict[str, Any]) -> None:
        self._service = service
        self._master_record = master_record
        self._asset_record = asset_record

        self._versions: Optional[Dict[VersionSize, AssetVersion]] = None

    ITEM_TYPES = {
        u"public.heic": AssetItemType.IMAGE,
        u"public.jpeg": AssetItemType.IMAGE,
        u"public.png": AssetItemType.IMAGE,
        u"com.apple.quicktime-movie": AssetItemType.MOVIE,
        u"com.adobe.raw-image": AssetItemType.IMAGE,
        u"com.canon.cr2-raw-image": AssetItemType.IMAGE,
        u'com.canon.crw-raw-image': AssetItemType.IMAGE,
        u'com.sony.arw-raw-image': AssetItemType.IMAGE,
        u'com.fuji.raw-image': AssetItemType.IMAGE,
        u'com.panasonic.rw2-raw-image': AssetItemType.IMAGE,
        u'com.nikon.nrw-raw-image': AssetItemType.IMAGE,
        u'com.pentax.raw-image': AssetItemType.IMAGE,
        u'com.nikon.raw-image': AssetItemType.IMAGE,
        u'com.olympus.raw-image': AssetItemType.IMAGE,
        u'com.canon.cr3-raw-image': AssetItemType.IMAGE,
        u'com.olympus.or-raw-image': AssetItemType.IMAGE,
    }

    ITEM_TYPE_EXTENSIONS = {
        u"public.heic": u"HEIC",
        u"public.jpeg": u"JPG",
        u"public.png": u"PNG",
        u"com.apple.quicktime-movie": u"MOV",
        u"com.adobe.raw-image": u"DNG",
        u"com.canon.cr2-raw-image": u"CR2",
        u'com.canon.crw-raw-image': u"CRW",
        u'com.sony.arw-raw-image': u"ARW",
        u'com.fuji.raw-image': u"RAF",
        u'com.panasonic.rw2-raw-image': u"RW2",
        u'com.nikon.nrw-raw-image': u"NRF",
        u'com.pentax.raw-image': u"PEF",
        u'com.nikon.raw-image': u"NEF",
        u'com.olympus.raw-image': u"ORF",
        u'com.canon.cr3-raw-image': u"CR3",
        u'com.olympus.or-raw-image': u"ORF",
    }

    PHOTO_VERSION_LOOKUP: Dict[VersionSize, str] = {
        AssetVersionSize.ORIGINAL: u"resOriginal",
        AssetVersionSize.ALTERNATIVE: u"resOriginalAlt",
        AssetVersionSize.MEDIUM: u"resJPEGMed",
        AssetVersionSize.THUMB: u"resJPEGThumb",
        AssetVersionSize.ADJUSTED: u"resJPEGFull",
        LivePhotoVersionSize.ORIGINAL: u"resOriginalVidCompl",
        LivePhotoVersionSize.MEDIUM: u"resVidMed",
        LivePhotoVersionSize.THUMB: u"resVidSmall",
    }

    VIDEO_VERSION_LOOKUP: Dict[VersionSize, str] = {
        AssetVersionSize.ORIGINAL: u"resOriginal",
        AssetVersionSize.MEDIUM: u"resVidMed",
        AssetVersionSize.THUMB: u"resVidSmall"
    }

    VERSION_FILENAME_SUFFIX_LOOKUP: Dict[VersionSize, str] = {
        AssetVersionSize.MEDIUM: u"medium",
        AssetVersionSize.THUMB: u"thumb",
        # LivePhotoVersionSize.MEDIUM: u"medium",
        # LivePhotoVersionSize.THUMB: u"thumb",
    }

    @property
    def id(self) -> str:
        return typing.cast(str, self._master_record['recordName'])

    @property
    def filename(self) -> str:
        fields = self._master_record['fields']
        if 'filenameEnc' in fields:
            filename_enc: Dict[str, Any] = fields['filenameEnc']
            def _get_value(input: Dict[str, Any]) -> str:
                return cast(str,input['value'])
            def _get_type(input: Dict[str, Any]) -> str:
                return cast(str, input['type'])
            
            def _match_type(string_parser: Callable[[str], str], base64_parser: Callable[[str], str]) -> Callable[[str], Callable[[str], str]]:
                def _internal(type:str) -> Callable[[str], str]:
                    if type == "STRING":
                        return string_parser
                    elif type == "ENCRYPTED_BYTES":
                        return base64_parser
                    else:
                        raise ValueError(f"Unsupported filename encoding {type}")
                return _internal

            parse_base64_value = compose(
                bytes_decode('utf-8'), 
                base64.b64decode,
            )
            
            parser_selector = compose(
                _match_type(identity, parse_base64_value),
                _get_type
            )

            type_parser = wrap_param_in_exception("Parsing filenameEnc type", parser_selector)

            _value_parser = type_parser(filename_enc)

            parse_and_clean = compose(
                self._service.filename_cleaner,
                _value_parser
            )

            extract_value_and_parse = compose(
                parse_and_clean,
                _get_value,
            )
            parser = wrap_param_in_exception("Parsing filenameEnc", extract_value_and_parse)

            _filename = parser(filename_enc)

            # _filename = self._service.filename_cleaner(base64.b64decode(
            #     fields['filenameEnc']['value']
            # ).decode('utf-8'))
            
            if self._service.file_match_policy == FileMatchPolicy.NAME_ID7:
                _a = base64.b64encode(self.id.encode('utf-8')).decode('ascii')[0:7]
                _filename = add_suffix_to_filename(f"_{_a}", _filename)
            return _filename

        # Some photos don't have a filename.
        # In that case, just use the truncated fingerprint (hash),
        # plus the correct extension.
        filename = re.sub('[^0-9a-zA-Z]', '_', self.id)[0:12]
        return '.'.join([filename, self.item_type_extension])

    @property
    def size(self) -> int:
        return typing.cast(int, self._master_record['fields']['resOriginalRes']['value']['size'])

    @property
    def created(self) -> datetime:
        return self.asset_date

    @property
    def asset_date(self) -> datetime:
        try:
            dt = datetime.fromtimestamp(
                self._asset_record['fields']['assetDate']['value'] / 1000.0,
                tz=pytz.utc)
        except:
            dt = datetime.fromtimestamp(0)
        return dt

    @property
    def added_date(self) -> datetime:
        dt = datetime.fromtimestamp(
            self._asset_record['fields']['addedDate']['value'] / 1000.0,
            tz=pytz.utc)
        return dt

    @property
    def dimensions(self) -> Tuple[int, int]:
        return (self._master_record['fields']['resOriginalWidth']['value'],
                self._master_record['fields']['resOriginalHeight']['value'])

    @property
    def item_type(self) -> Optional[AssetItemType]:
        fields = self._master_record['fields']
        if 'itemType' not in fields:
            # raise ValueError(f"Cannot find itemType in {fields!r}")
            return None
        item_type_field = fields['itemType']
        if 'value' not in item_type_field:
            # raise ValueError(f"Cannot find value in itemType {item_type_field!r}")
            return None
        item_type = item_type_field['value']
        if item_type in self.ITEM_TYPES:
            return self.ITEM_TYPES[item_type]
        if self.filename.lower().endswith(('.heic', '.png', '.jpg', '.jpeg')):
            return AssetItemType.IMAGE
        return AssetItemType.MOVIE

    @property
    def item_type_extension(self) -> str:
        fields = self._master_record['fields']
        if 'itemType' not in fields or 'value' not in fields['itemType']:
            return 'unknown'
        item_type = self._master_record['fields']['itemType']['value']
        if item_type in self.ITEM_TYPE_EXTENSIONS:
            return self.ITEM_TYPE_EXTENSIONS[item_type]
        return 'unknown'

    @property
    def versions(self) -> Dict[VersionSize, AssetVersion]:
        if not self._versions:
            _versions: Dict[VersionSize, AssetVersion] = {}
            if self.item_type == AssetItemType.MOVIE:
                typed_version_lookup: Dict[VersionSize, str] = self.VIDEO_VERSION_LOOKUP
            else:
                typed_version_lookup = self.PHOTO_VERSION_LOOKUP

            # self._master_record["dummy"] ## to trigger dump

            for key, prefix in typed_version_lookup.items():
                f: Optional[Dict[str, Any]] = None
                if '%sRes' % prefix in self._asset_record['fields']:
                    f = self._asset_record['fields']
                if not f and '%sRes' % prefix in self._master_record['fields']:
                    f = self._master_record['fields']
                if f:
                    version: Dict[str, Any] = {'filename': self.filename}

                    # width_entry = f.get('%sWidth' % prefix)
                    # if width_entry:
                    #     version['width'] = width_entry['value']
                    # else:
                    #     version['width'] = None

                    # height_entry = f.get('%sHeight' % prefix)
                    # if height_entry:
                    #     version['height'] = height_entry['value']
                    # else:
                    #     version['height'] = None

                    size_entry = f.get('%sRes' % prefix)
                    if size_entry:
                        version['size'] = size_entry['value']['size']
                        version['url'] = size_entry['value']['downloadURL']
                    else:
                        raise ValueError(f"Expected {prefix}Res, but missing it")
                        # version['size'] = None
                        # version['url'] = None

                    type_entry = f.get('%sFileType' % prefix)
                    if type_entry:
                        version['type'] = type_entry['value']
                    else:
                        raise ValueError(f"Expected {prefix}FileType, but missing it")
                        # version['type'] = None

                    # Change live photo movie file extension to .MOV
                    if ((self.item_type or AssetItemType.IMAGE) == AssetItemType.IMAGE and
                        version['type'] == "com.apple.quicktime-movie"):
                        version['filename'] = self._service.lp_filename_generator(self.filename) # without size
                    else:
                        # for non live photo movie, try to change file type to match asset type
                        _f, _e = os.path.splitext(version["filename"])
                        version["filename"] = _f + "." + self.ITEM_TYPE_EXTENSIONS.get(version["type"], _e[1:])

                    # add size suffix
                    if key in self.VERSION_FILENAME_SUFFIX_LOOKUP:
                        _size_suffix = self.VERSION_FILENAME_SUFFIX_LOOKUP[key]
                        version["filename"] = add_suffix_to_filename(f"-{_size_suffix}", version["filename"])

                    _versions[key] = AssetVersion(version["filename"], version['size'], version['url'], version['type'])

            # swap original & alternative according to swap_raw_policy
            if AssetVersionSize.ALTERNATIVE in _versions and (("raw" in _versions[AssetVersionSize.ALTERNATIVE].type and self._service.raw_policy == RawTreatmentPolicy.AS_ORIGINAL) or ("raw" in _versions[AssetVersionSize.ORIGINAL].type and self._service.raw_policy == RawTreatmentPolicy.AS_ALTERNATIVE)):
                _a = copy.copy(_versions[AssetVersionSize.ALTERNATIVE])
                _o = copy.copy(_versions[AssetVersionSize.ORIGINAL])
                _versions[AssetVersionSize.ALTERNATIVE] = _o
                _versions[AssetVersionSize.ORIGINAL] = _a

            self._versions = _versions

        return self._versions

    def download(self, url: str) -> Response:
        return self._service.session.get(
            url,
            stream=True
        )

    def __repr__(self) -> str:
        return "<%s: id=%s>" % (
            type(self).__name__,
            self.id
        )



# mypy: ignore-errors
from __future__ import absolute_import
from datetime import datetime, timedelta
from calendar import monthrange
import time

from tzlocal import get_localzone


class CalendarService(object):
    """
    The 'Calendar' iCloud service, connects to iCloud and returns events.
    """
    def __init__(self, service_root, session, params):
        self.session = session
        self.params = params
        self._service_root = service_root
        self._calendar_endpoint = '%s/ca' % self._service_root
        self._calendar_refresh_url = '%s/events' % self._calendar_endpoint
        self._calendar_event_detail_url = '%s/eventdetail' % (
            self._calendar_endpoint,
        )

    def get_event_detail(self, pguid, guid):
        """
        Fetches a single event's details by specifying a pguid
        (a calendar) and a guid (an event's ID).
        """
        params = dict(self.params)
        params.update({'lang': 'en-us', 'usertz': get_localzone().zone})
        url = '%s/%s/%s' % (self._calendar_event_detail_url, pguid, guid)
        req = self.session.get(url, params=params)
        self.response = req.json()
        return self.response['Event'][0]

    def refresh_client(self, from_dt=None, to_dt=None):
        """
        Refreshes the CalendarService endpoint, ensuring that the
        event data is up-to-date. If no 'from_dt' or 'to_dt' datetimes
        have been given, the range becomes this month.
        """
        today = datetime.today()
        first_day, last_day = monthrange(today.year, today.month)
        if not from_dt:
            from_dt = datetime(today.year, today.month, first_day)
        if not to_dt:
            to_dt = datetime(today.year, today.month, last_day)
        params = dict(self.params)
        params.update({
            'lang': 'en-us',
            'usertz': get_localzone().zone,
            'startDate': from_dt.strftime('%Y-%m-%d'),
            'endDate': to_dt.strftime('%Y-%m-%d')
        })
        req = self.session.get(self._calendar_refresh_url, params=params)
        self.response = req.json()

    def events(self, from_dt=None, to_dt=None):
        """
        Retrieves events for a given date range, by default, this month.
        """
        self.refresh_client(from_dt, to_dt)
        return self.response['Event']



from enum import Enum
from typing import Union


class AssetVersionSize(Enum):
    ORIGINAL = "original"
    ADJUSTED = "adjusted"
    ALTERNATIVE = "alternative"
    MEDIUM = "medium"
    THUMB = "thumb"

    def __str__(self) -> str:
        return self.name


class LivePhotoVersionSize(Enum):
    ORIGINAL = "originalVideo"
    MEDIUM = "mediumVideo"
    THUMB = "smallVideo"

    def __str__(self) -> str:
        return self.name


VersionSize = Union[AssetVersionSize, LivePhotoVersionSize]



from foundation.core import compose, identity
from icloudpd.paths import clean_filename, remove_unicode_chars
from icloudpd.base import lp_filename_concatinator, lp_filename_original, download_builder

from pyicloud_ipd.file_match import FileMatchPolicy
from pyicloud_ipd.base import PyiCloudService
from icloudpd_web.api.data_models import PolicyConfigs

from typing import Callable

from inspect import signature


class ICloudManager:
    def __init__(self):
        self._icloud_instances: dict[str, PyiCloudService] = {}

    def get_instance(self, username: str) -> PyiCloudService | None:
        """
        Get the instance for the given username.
        """
        return self._icloud_instances.get(username)

    def set_instance(self, username: str, instance: PyiCloudService):
        """
        Set the instance for the given username.
        """
        assert (
            self._icloud_instances.get(username) is None
        ), "Trying to set an icloud instance that already exists"
        self._icloud_instances[username] = instance

    def update_instance(self, username: str, attributes: dict):
        """
        Update the attributes of the instance with the given username.
        """
        instance = self._icloud_instances.get(username)
        assert instance is not None, "Trying to update non-existing instance"
        for key in attributes:
            assert hasattr(instance, key), f"Instance does not have attribute '{key}'"
        for key, value in attributes.items():
            setattr(instance, key, value)

    def remove_instances(self, active_usernames: list[str]):
        """
        Remove all instances that are not in the list of active usernames.
        """
        for username in self._icloud_instances:
            if username not in active_usernames:
                self._icloud_instances.pop(username, None)


def request_2sa(icloud: PyiCloudService) -> None:
    """
    Request 2SA code using SMS from the first trusted device.
    Difference between 2FA and 2SA: https://discussions.apple.com/thread/7932277
    """
    devices = icloud.trusted_devices
    if len(devices) == 0:
        raise ValueError("No devices available for 2SA")
    # Request 2SA from the first trusteddevice
    phone_number = devices[0]["phoneNumber"]
    print(f"Requesting 2SA code via SMS to {phone_number}")
    icloud.send_verification_code(devices[0])


def build_downloader_builder_args(configs: PolicyConfigs) -> dict:
    downloader_args = {
        "only_print_filenames": False,
        "dry_run": False,
        **configs.model_dump(),
    }
    builder_params = signature(download_builder).parameters.keys()
    downloader_args = {k: v for k, v in downloader_args.items() if k in builder_params}
    # Map size to primary_sizes
    downloader_args["primary_sizes"] = configs.size
    return downloader_args


def build_filename_cleaner(keep_unicode_in_filenames: bool) -> Callable[[str], str]:
    """Map keep_unicode parameter for function for cleaning filenames"""
    return compose(
        (remove_unicode_chars if not keep_unicode_in_filenames else identity),
        clean_filename,
    )


def build_lp_filename_generator(live_photo_mov_filename_policy: str) -> Callable[[str], str]:
    return (
        lp_filename_original
        if live_photo_mov_filename_policy == "original"
        else lp_filename_concatinator
    )


def file_match_policy_generator(policy: str) -> FileMatchPolicy:
    match policy:
        case "name-size-dedup-with-suffix":
            return FileMatchPolicy.NAME_SIZE_DEDUP_WITH_SUFFIX
        case "name-id7":
            return FileMatchPolicy.NAME_ID7
        case _:
            raise ValueError(f"policy was provided with unsupported value of '{policy}'")


def build_raw_policy(align_raw: str) -> str:
    match align_raw:
        case "original":
            return "as-original"
        case "alternative":
            return "as-alternative"
        case "as-is":
            return "as-is"
        case _:
            raise ValueError(f"align_raw was provided with unsupported value of '{align_raw}'")


def build_pyicloudservice_args(configs: PolicyConfigs) -> dict:
    return {
        "filename_cleaner": build_filename_cleaner(configs.keep_unicode_in_filenames),
        "lp_filename_generator": build_lp_filename_generator(
            configs.live_photo_mov_filename_policy
        ),
        "raw_policy": build_raw_policy(configs.align_raw),
        "file_match_policy": file_match_policy_generator(configs.file_match_policy),
    }



# mypy: ignore-errors
import sys

import six

from pyicloud_ipd.utils import underscore_to_camelcase


class AccountService(object):
    def __init__(self, service_root, session, params):
        self.session = session
        self.params = params
        self._service_root = service_root
        self._devices = []

        self._acc_endpoint = '%s/setup/web/device' % self._service_root
        self._account_devices_url = '%s/getDevices' % self._acc_endpoint

        req = self.session.get(self._account_devices_url, params=self.params)
        self.response = req.json()

        for device_info in self.response['devices']:
            # device_id = device_info['udid']
            # self._devices[device_id] = AccountDevice(device_info)
            self._devices.append(AccountDevice(device_info))

    @property
    def devices(self):
        return self._devices


@six.python_2_unicode_compatible
class AccountDevice(dict): # type: ignore[type-arg]
    def __init__(self, device_info):
        super(AccountDevice, self).__init__(device_info)

    def __getattr__(self, name):
        try:
            return self[underscore_to_camelcase(name)]
        except KeyError:
            raise AttributeError(name)

    def __str__(self):
        return u"{display_name}: {name}".format(
            display_name=self.model_display_name,
            name=self.name,
        )

    def __repr__(self):
        return '<{display}>'.format(
            display=(
                six.text_type(self)
                if sys.version_info[0] >= 3 else
                six.text_type(self).encode('utf8', 'replace')
            )
        )



import logging
import io
from typing import Tuple, Callable
from pyicloud_ipd.base import PyiCloudService
from icloudpd.base import (
    compose_handlers,
    session_error_handle_builder,
    internal_error_handle_builder,
)


def build_logger_level(level: str) -> int:
    match level:
        case "debug":
            return logging.DEBUG
        case "info":
            return logging.INFO
        case "error":
            return logging.ERROR
        case _:
            raise ValueError(f"Unsupported logger level: {level}")


class LogCaptureStream(io.StringIO):
    def __init__(self):
        super().__init__()
        self.buffer: list[str] = []

    def write(self, message):
        # Store each log message in the buffer
        self.buffer.append(message)
        super().write(message)

    def read_new_lines(self) -> list[str]:
        # Return new lines and clear the buffer
        if self.buffer:
            new_lines = "".join(self.buffer)
            self.buffer = []
            return new_lines
        return ""


def build_logger(policy_name: str) -> Tuple[logging.Logger, LogCaptureStream]:
    log_capture_stream = LogCaptureStream()
    logger = logging.getLogger(f"{policy_name}-logger")
    logger.handlers.clear()
    stream_handler = logging.StreamHandler(log_capture_stream)
    formatter = logging.Formatter(
        fmt="%(asctime)s %(levelname)-8s %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    return logger, log_capture_stream


def build_photos_exception_handler(logger: logging.Logger, icloud: PyiCloudService) -> Callable:
    session_exception_handler = session_error_handle_builder(logger, icloud)
    internal_error_handler = internal_error_handle_builder(logger)

    error_handler = compose_handlers([session_exception_handler, internal_error_handler])
    return error_handler



# mypy: ignore-errors
from __future__ import absolute_import
from datetime import datetime, timedelta
import time
import uuid
import json

from tzlocal import get_localzone


class RemindersService(object):
    def __init__(self, service_root, session, params):
        self.session = session
        self.params = params
        self._service_root = service_root
        self.lists = {}
        self.collections = {}

        self.refresh()

    def refresh(self):
        params_reminders = dict(self.params)
        params_reminders.update({
            'clientVersion': '4.0',
            'lang': 'en-us',
            'usertz': get_localzone().zone
        })

        # Open reminders
        req = self.session.get(
            self._service_root + '/rd/startup',
            params=params_reminders
        )

        startup = req.json()

        self.lists = {}
        self.collections = {}
        for collection in startup['Collections']:
            temp = []
            self.collections[collection['title']] = {
                'guid': collection['guid'],
                'ctag': collection['ctag']
            }
            for reminder in startup['Reminders']:

                if reminder['pGuid'] != collection['guid']:
                    continue
                if 'dueDate' in reminder:
                    if reminder['dueDate']:
                        due = datetime(
                            reminder['dueDate'][1],
                            reminder['dueDate'][2], reminder['dueDate'][3],
                            reminder['dueDate'][4], reminder['dueDate'][5]
                        )
                    else:
                        due = None
                else:
                    due = None
                if reminder['description']:
                    desc = reminder['description']
                else:
                    desc = ""
                temp.append({
                    "title": reminder['title'],
                    "desc": desc,
                    "due": due
                })
            self.lists[collection['title']] = temp

    def post(self, title, description="", collection=None, dueDate=None):
        pguid = 'tasks'
        if collection:
            if collection in self.collections:
                pguid = self.collections[collection]['guid']

        params_reminders = dict(self.params)
        params_reminders.update({
            'clientVersion': '4.0',
            'lang': 'en-us',
            'usertz': get_localzone().zone
        })

        dueDateList = None
        if dueDate:
            dueDateList = [
                int(str(dueDate.year) + str(dueDate.month) + str(dueDate.day)),
                dueDate.year,
                dueDate.month,
                dueDate.day,
                dueDate.hour,
                dueDate.minute
            ]

        req = self.session.post(
            self._service_root + '/rd/reminders/tasks',
            data=json.dumps({
                "Reminders": {
                    'title': title,
                    "description": description,
                    "pGuid": pguid,
                    "etag": None,
                    "order": None,
                    "priority": 0,
                    "recurrence": None,
                    "alarms": [],
                    "startDate": None,
                    "startDateTz": None,
                    "startDateIsAllDay": False,
                    "completedDate": None,
                    "dueDate": dueDateList,
                    "dueDateIsAllDay": False,
                    "lastModifiedDate": None,
                    "createdDate": None,
                    "isFamily": None,
                    "createdDateExtended": int(time.time()*1000),
                    "guid": str(uuid.uuid4())
                },
                "ClientState": {"Collections": list(self.collections.values())}
            }),
            params=params_reminders)
        return req.ok



from fastapi import WebSocket, WebSocketDisconnect
from typing import Dict, Set
import json
import logging

logger = logging.getLogger(__name__)


class ConnectionManager:
    def __init__(self):
        self.active_connections: Dict[str, Set[WebSocket]] = {}

    async def connect(self, websocket: WebSocket, client_id: str):
        await websocket.accept()
        if client_id not in self.active_connections:
            self.active_connections[client_id] = set()
        self.active_connections[client_id].add(websocket)
        logger.info(f"Client {client_id} connected")

    def disconnect(self, websocket: WebSocket, client_id: str):
        if client_id in self.active_connections:
            self.active_connections[client_id].discard(websocket)
            if not self.active_connections[client_id]:
                del self.active_connections[client_id]
        logger.info(f"Client {client_id} disconnected")

    async def send_message(self, message: dict, client_id: str):
        if client_id not in self.active_connections:
            return

        for connection in self.active_connections[client_id]:
            try:
                await connection.send_json(message)
            except Exception as e:
                logger.error(f"Error sending message to client {client_id}: {e}")


manager = ConnectionManager()


async def handle_websocket(websocket: WebSocket, client_id: str):
    try:
        await manager.connect(websocket, client_id)
        while True:
            try:
                data = await websocket.receive_text()
                message = json.loads(data)
                # Handle incoming messages here
                await manager.send_message(
                    {"type": "acknowledgment", "data": message}, client_id
                )
            except json.JSONDecodeError:
                await manager.send_message(
                    {"type": "error", "message": "Invalid JSON format"}, client_id
                )
    except WebSocketDisconnect:
        manager.disconnect(websocket, client_id)
    except Exception as e:
        logger.error(f"WebSocket error: {e}")
        manager.disconnect(websocket, client_id)



import json
import sys
from typing import Any, Dict, Iterable, Optional, Sequence
import typing

import six

from pyicloud_ipd.exceptions import PyiCloudNoDevicesException


class FindMyiPhoneServiceManager(object):
    """ The 'Find my iPhone' iCloud service

    This connects to iCloud and return phone data including the near-realtime
    latitude and longitude.

    """

    def __init__(self, service_root: str, session:Any, params: Dict[str, Any]):
        self.session = session
        self.params = params
        self._service_root = service_root
        self._fmip_endpoint = '%s/fmipservice/client/web' % self._service_root
        self._fmip_refresh_url = '%s/refreshClient' % self._fmip_endpoint
        self._fmip_sound_url = '%s/playSound' % self._fmip_endpoint
        self._fmip_message_url = '%s/sendMessage' % self._fmip_endpoint
        self._fmip_lost_url = '%s/lostDevice' % self._fmip_endpoint

        self._devices:Dict[str, AppleDevice] = {}
        self.refresh_client()

    def refresh_client(self) -> None:
        """ Refreshes the FindMyiPhoneService endpoint,

        This ensures that the location data is up-to-date.

        """
        req = self.session.post(
            self._fmip_refresh_url,
            params=self.params,
            data=json.dumps(
                {
                    'clientContext': {
                        'fmly': True,
                        'shouldLocate': True,
                        'selectedDevice': 'all',
                    }
                }
            )
        )
        self.response = req.json()

        for device_info in self.response['content']:
            device_id = device_info['id']
            if device_id not in self._devices:
                self._devices[device_id] = AppleDevice(
                    device_info,
                    self.session,
                    self.params,
                    manager=self,
                    sound_url=self._fmip_sound_url,
                    lost_url=self._fmip_lost_url,
                    message_url=self._fmip_message_url,
                )
            else:
                self._devices[device_id].update(device_info)

        if not self._devices:
            raise PyiCloudNoDevicesException()

    def __getitem__(self, key: Any) -> "AppleDevice":
        if isinstance(key, int):
            if six.PY3:
                key = list(self.keys())[key] # type: ignore[operator]
            else:
                key = self.keys()[key]
        return self._devices[key]

    def __getattr__(self, attr: str) -> "AppleDevice":
        return typing.cast(AppleDevice, getattr(self._devices, attr))

    def __unicode__(self) -> str:
        return six.text_type(self._devices)

    def __str__(self) -> str:
        as_unicode = self.__unicode__()
        if sys.version_info[0] >= 3:
            return as_unicode
        else:
            return as_unicode.encode('ascii', 'ignore')

    def __repr__(self) -> str:
        return six.text_type(self)
    

class AppleDevice(object):
    def __init__(
        self, content:Dict[str,Any], session:Any, params:Dict[str, Any], manager: FindMyiPhoneServiceManager,
        sound_url:Optional[str]=None, lost_url:Optional[str]=None, message_url:Optional[str]=None
    ):
        self.content = content
        self.manager = manager
        self.session = session
        self.params = params

        self.sound_url = sound_url
        self.lost_url = lost_url
        self.message_url = message_url

    def update(self, data: Dict[str, Any]) -> None:
        self.content = data

    def location(self) -> Any:
        self.manager.refresh_client()
        return self.content['location']

    def status(self, additional:Sequence[str]=[]) -> Dict[str, Any]:
        """ Returns status information for device.

        This returns only a subset of possible properties.
        """
        self.manager.refresh_client()
        fields = ['batteryLevel', 'deviceDisplayName', 'deviceStatus', 'name']
        fields += additional
        properties: Dict[str, Any] = {}
        for field in fields:
            properties[field] = self.content.get(field)
        return properties

    def play_sound(self, subject:str='Find My iPhone Alert') -> None:
        """ Send a request to the device to play a sound.

        It's possible to pass a custom message by changing the `subject`.
        """
        data = json.dumps({
            'device': self.content['id'],
            'subject': subject,
            'clientContext': {
                'fmly': True
            }
        })
        self.session.post(
            self.sound_url,
            params=self.params,
            data=data
        )

    def display_message(
        self, subject:str='Find My iPhone Alert', message:str="This is a note",
        sounds:bool=False
    ) -> None:
        """ Send a request to the device to play a sound.

        It's possible to pass a custom message by changing the `subject`.
        """
        data = json.dumps(
            {
                'device': self.content['id'],
                'subject': subject,
                'sound': sounds,
                'userText': True,
                'text': message
            }
        )
        self.session.post(
            self.message_url,
            params=self.params,
            data=data
        )

    def lost_device(
        self, number:str,
        text:str='This iPhone has been lost. Please call me.',
        newpasscode:str=""
    ) -> None:
        """ Send a request to the device to trigger 'lost mode'.

        The device will show the message in `text`, and if a number has
        been passed, then the person holding the device can call
        the number without entering the passcode.
        """
        data = json.dumps({
            'text': text,
            'userText': True,
            'ownerNbr': number,
            'lostModeEnabled': True,
            'trackingEnabled': True,
            'device': self.content['id'],
            'passcode': newpasscode
        })
        self.session.post(
            self.lost_url,
            params=self.params,
            data=data
        )

    @property
    def data(self) -> Dict[str, Any]:
        return self.content

    def __getitem__(self, key: str) -> Any:
        return self.content[key]

    def __getattr__(self, attr: str) -> Any:
        return getattr(self.content, attr)

    def __unicode__(self) -> str:
        display_name = self['deviceDisplayName']
        name = self['name']
        return '%s: %s' % (
            display_name,
            name,
        )

    def __str__(self) -> str:
        as_unicode = self.__unicode__()
        if sys.version_info[0] >= 3:
            return as_unicode
        else:
            return as_unicode.encode('ascii', 'ignore')

    def __repr__(self) -> str:
        return '<AppleDevice(%s)>' % str(self)



import datetime
from typing import Callable, NamedTuple, TypeVar

import pytz
from tzlocal import get_localzone
from . import core


class VersionInfo(NamedTuple):
    version: str
    commit_sha: str
    commit_timestamp: int


# will be updated by CI
version_info = VersionInfo(
    version="2025.01.09",
    commit_sha="abcdefgh",
    commit_timestamp=1234567890,
)


def version_info_formatted() -> str:
    vi = version_info
    ts = datetime.datetime.fromtimestamp(vi.commit_timestamp, tz=pytz.utc).astimezone(
        get_localzone()
    )
    return f"version:{vi.version}, commit sha:{vi.commit_sha}, commit timestamp:{ts:%c %Z}"


def bytes_decode(encoding: str) -> Callable[[bytes], str]:
    def _internal(inp: bytes) -> str:
        return inp.decode(encoding)

    return _internal


T_in = TypeVar("T_in")
T_out = TypeVar("T_out")


def wrap_param_in_exception(caption: str, func: Callable[[T_in], T_out]) -> Callable[[T_in], T_out]:
    """Decorator to wrap excpetion into ValueError with the dump of the input parameter"""

    def _internal(input: T_in) -> T_out:
        try:
            return func(input)
        except Exception as err:
            raise ValueError(f"Invalid Input ({caption}): {input!r}") from err

    return _internal


__all__ = [
    'VersionInfo',
    'version_info',
    'version_info_formatted',
    'bytes_decode',
    'wrap_param_in_exception',
    'core'  # Expose the core module
]



from icloudpd_web.api.icloud_utils import ICloudManager
from icloudpd_web.api.policy_handler import PolicyHandler, PolicyStatus
from icloudpd_web.api.data_models import NON_POLICY_FIELDS

import toml
import os


class SessionHandler:
    @property
    def policies(self) -> list[dict]:
        return [policy.dump() for policy in self._policies]

    @property
    def policy_names(self) -> list[str]:
        return [policy.name for policy in self._policies]

    def __init__(self, saved_policies_path: str):
        self._policies: list[PolicyHandler] = []
        self._saved_policies_path: str = saved_policies_path
        self._icloud_manager = ICloudManager()
        self._load_policies()

    def _load_policies_from_toml(self, saved_policies: list[dict]) -> list[dict]:
        for policy in saved_policies:
            assert "name" in policy, "Policy must have a name"
            assert "username" in policy, "Policy must have a username"
            assert "directory" in policy, "Policy must have a directory"
            assert policy["name"] not in self.policy_names, "Policy name must be unique"
            self._policies.append(PolicyHandler(icloud_manager=self._icloud_manager, **policy))

    def _load_policies(self):
        """
        Load the policies from the file if it exists.
        """
        if os.path.exists(self._saved_policies_path):
            with open(self._saved_policies_path, "r") as file:
                saved_policies = toml.load(file).get("policy", [])
                self._load_policies_from_toml(saved_policies)

    def _save_policies(self):
        """
        Save the policies to a toml file at the given path.
        """
        with open(self._saved_policies_path, "w") as file:
            policies_to_save = {
                "policy": [policy.dump(excludes=NON_POLICY_FIELDS) for policy in self._policies]
            }
            toml.dump(policies_to_save, file)

    def dump_policies_as_toml(self) -> str:
        """
        Dump the policies as a TOML string.
        """
        return toml.dumps(
            {"policy": [policy.dump(excludes=NON_POLICY_FIELDS) for policy in self._policies]}
        )

    def get_policy(self, name: str) -> PolicyHandler | None:
        """
        Return the policy with the given name.
        """
        for policy in self._policies:
            if policy.name == name:
                return policy
        return None

    def save_policy(self, policy_name: str, **kwargs):
        """
        Update the parameters of an existing policy.
        If policy with new name exists, update that policy.
        """
        assert policy_name in [policy.name for policy in self._policies], "Policy does not exist"
        form_name = kwargs.get("name", "")
        if form_name == policy_name:  # name is not changed
            self.get_policy(policy_name).update(config_updates=kwargs)
        else:  # name is changed
            assert (
                form_name not in self.policy_names
            ), f"Policy with name {form_name} already exists"
            policy = self.get_policy(policy_name)
            policy.name = form_name
            policy.update(config_updates=kwargs)

        self._save_policies()

    def create_policy(self, **kwargs):
        """
        Create a new policy with the given parameters.
        """
        policy_name = kwargs.get("name")
        assert policy_name, "Policy name must be provided"
        assert (
            policy_name not in self.policy_names
        ), f"Policy with name {policy_name} already exists"
        self._policies.append(PolicyHandler(icloud_manager=self._icloud_manager, **kwargs))
        self._save_policies()

    def delete_policy(self, policy_name: str):
        """
        Delete the policy with the given name.
        """
        assert policy_name in self.policy_names, f"Policy {policy_name} does not exist"
        self._policies = [policy for policy in self._policies if policy.name != policy_name]
        self._save_policies()

    def replace_policies(self, toml_content: str):
        """
        Replace the current policies with the policies defined in the list of dictionaries.
        """
        self._policies = []
        self._icloud_manager = ICloudManager()
        read_policies = toml.loads(toml_content).get("policy", [])
        self._load_policies_from_toml(read_policies)
        self._save_policies()

    def icloud_instance_occupied_by(self, username: str) -> str | None:
        """
        Check if another policy using the same username is running.
        """
        for policy in self._policies:
            if policy.username == username and policy.status == PolicyStatus.RUNNING:
                return policy.name
        return None



"""Get/set EXIF dates from photos"""

import logging
import typing

import piexif
from piexif._exceptions import InvalidImageDataError


def get_photo_exif(logger: logging.Logger, path: str) -> typing.Optional[str]:
    """Get EXIF date for a photo, return nothing if there is an error"""
    try:
        exif_dict: piexif.ExifIFD = piexif.load(path)
        return typing.cast(typing.Optional[str], exif_dict.get("Exif").get(36867))
    except (ValueError, InvalidImageDataError):
        logger.debug("Error fetching EXIF data for %s", path)
        return None


def set_photo_exif(logger: logging.Logger, path: str, date: str) -> None:
    """Set EXIF date on a photo, do nothing if there is an error"""
    try:
        exif_dict = piexif.load(path)
        exif_dict.get("1st")[306] = date
        exif_dict.get("Exif")[36867] = date
        exif_dict.get("Exif")[36868] = date
        exif_bytes = piexif.dump(exif_dict)
        piexif.insert(exif_bytes, path)
    except (ValueError, InvalidImageDataError):
        logger.debug("Error setting EXIF data for %s", path)
        return



from typing import Union


class AssetVersion:
    def __init__(self, filename: str, size: int, url: str, type: str) -> None:
        self.filename = filename
        self.size = size
        self.url = url
        self.type = type

    def __eq__(self, other: object) -> bool: 
        if not isinstance(other, AssetVersion):
            # don't attempt to compare against unrelated types
            return NotImplemented
        return self.filename == other.filename and self.size == other.size and self.url == other.url and self.type == other.type


from icloudpd_web.app import create_app
from importlib import resources
import socketio

# Get the path to the static files
STATIC_DIR = str(resources.files("icloudpd_web")) + "/webapp"

# Create the app with static file serving enabled
app, sio = create_app(serve_static=True, static_dir=STATIC_DIR)

# ASGI app by wrapping FastAPI with Socket.IO
socket_app = socketio.ASGIApp(sio, app)



from icloudpd_web.api.data_models import PolicyConfigs, AuthenticationResult
from icloudpd_web.api.icloud_utils import (
    build_pyicloudservice_args,
    build_downloader_builder_args,
    request_2sa,
    ICloudManager,
)
from icloudpd_web.api.logger import build_logger_level, build_photos_exception_handler
from icloudpd_web.api.download_option_utils import (
    handle_recent_until_found,
    log_at_download_start,
    should_break,
    check_folder_structure,
)
from pyicloud_ipd.base import PyiCloudService
from pyicloud_ipd.exceptions import PyiCloudFailedLoginException
from pyicloud_ipd.services.photos import PhotoAlbum
from icloudpd.base import download_builder, delete_photo, retrier
from icloudpd.counter import Counter
from icloudpd.autodelete import autodelete_photos

from enum import Enum
from typing import cast, Callable
from functools import partial
import asyncio
import logging
import os


class PolicyStatus(Enum):
    RUNNING = "running"
    STOPPED = "stopped"
    ERRORED = "errored"


class PolicyHandler:
    @property
    def name(self) -> str:
        return self._name

    @name.setter
    def name(self, value: str):
        assert isinstance(value, str), "Policy name must be a string"
        self._name = value

    @property
    def status(self) -> PolicyStatus:
        return self._status

    @status.setter
    def status(self, value: PolicyStatus):
        assert isinstance(value, PolicyStatus), "Status must be a PolicyStatus"
        self._status = value

    @property
    def progress(self) -> int:
        return self._progress

    @progress.setter
    def progress(self, value: int):
        assert isinstance(value, int), "Progress must be an integer"
        assert 0 <= value <= 100, "Progress must be between 0 and 100"
        self._progress = value

    @property
    def authenticated(self) -> bool:
        return (
            self.icloud is not None
            and not self.icloud.requires_2sa
            and not self.icloud.requires_2fa
        )

    @property
    def albums(self) -> list[str]:
        """
        Return a list of all albums available to the user.
        """
        if not self.authenticated:
            return []
        if library := self.library_name:
            return [str(a) for a in self.icloud.photos.libraries[library].albums.values()]
        else:
            return []

    @property
    def library_name(self) -> str:
        """
        Find the actual library name from icloud given the library name in the configs.
        """
        assert self.authenticated, "Can only get library name when authenticated"
        libraries = list(self.icloud.photos.libraries.keys())
        shared_library_name = next((lib for lib in libraries if "SharedSync" in lib), None)
        if shared_library_name and self._configs.library == "Shared Library":
            return shared_library_name
        elif self._configs.library == "Personal Library":
            return "PrimarySync"
        else:
            return None

    @property
    def icloud(self) -> PyiCloudService:
        return self._icloud_manager.get_instance(self.username)

    @icloud.setter
    def icloud(self, instance: PyiCloudService):
        self._icloud_manager.set_instance(self.username, instance)

    @property
    def username(self) -> str:
        return self._configs.username

    def __init__(self, name: str, icloud_manager: ICloudManager, **kwargs):
        self._name = name
        self._configs = PolicyConfigs(**kwargs)  # validate the configs and fill-in defaults
        self._status = PolicyStatus.STOPPED
        self._progress = 0
        self._icloud_manager = icloud_manager

    def dump(self, excludes: list[str] = []) -> dict:
        policy_dict = {
            "name": self._name,
            "status": self._status.value,
            "progress": self._progress,
            "authenticated": self.authenticated,
            "albums": self.albums,
            **self._configs.model_dump(),
        }
        for exclude in excludes:
            policy_dict.pop(exclude, None)
        return policy_dict

    def update(self, config_updates: dict):
        """
        Update the policy configs. Should only be called when status is STOPPED.
        """
        assert (
            self._status == PolicyStatus.STOPPED or self._status == PolicyStatus.ERRORED
        ), "Can only update policy when policy is stopped or errored"
        new_config_args = self._configs.model_dump()
        new_config_args.update(config_updates)
        self._configs = PolicyConfigs(**new_config_args)
        self._progress = 0

    def authenticate(self, password: str):
        """
        Create the icloud instance with the given password. User may need to provide MFA code to finish authentication.
        """
        assert self._status == PolicyStatus.STOPPED, "Can only authenticate when policy is stopped"
        assert not self.authenticated, "Can only authenticate when it is not authenticated"
        try:
            pyicloudservice_args = build_pyicloudservice_args(self._configs)
            self.icloud = PyiCloudService(
                **pyicloudservice_args,
                domain=self._configs.domain,
                apple_id=self.username,
                password=password,
            )
        except PyiCloudFailedLoginException as e:
            return AuthenticationResult.FAILED, e.args[0]
        if self.authenticated:
            return AuthenticationResult.SUCCESS, "Authenticated."
        else:
            if (
                self.icloud.requires_2sa and not self.icloud.requires_2fa
            ):  # User does not have MFA enabled, request 2SA using SMS manually
                request_2sa(self.icloud)
            return AuthenticationResult.MFA_REQUIRED, "MFA required."

    def provide_mfa(self, mfa_code: str):
        """
        Provide the MFA code to the icloud instance to finish authentication.
        """
        assert not self.authenticated, "Can only provide MFA when policy is not authenticated"
        self.icloud.validate_2fa_code(mfa_code)
        if not self.authenticated:
            return AuthenticationResult.MFA_REQUIRED, "Wrong MFA code."
        else:
            return AuthenticationResult.SUCCESS, "Authenticated."

    async def start(self, logger: logging.Logger):
        """
        Start running the policy for download.
        """
        assert self.authenticated, "Can only start when authenticated"
        self._status = PolicyStatus.RUNNING
        logger.setLevel(build_logger_level(self._configs.log_level))

        try:
            logger.info(f"Starting policy: {self._name}...")

            pyicloudservice_args = build_pyicloudservice_args(self._configs)
            self._icloud_manager.update_instance(
                username=self.username,
                attributes=pyicloudservice_args,
            )

            download_photo: Callable = download_builder(
                logger=logger, **build_downloader_builder_args(self._configs)
            )(self.icloud)

            async def async_download_photo(*args, **kwargs):
                loop = asyncio.get_running_loop()
                return await loop.run_in_executor(None, download_photo, *args, **kwargs)

            directory = os.path.normpath(cast(str, self._configs.directory))
            check_folder_structure(logger, directory, self._configs.folder_structure)

            if (library_name := self.library_name) is None:
                raise ValueError(f"Unavailable library: {self._configs.library}")
            library = self.icloud.photos.libraries[library_name]
            assert (
                self._configs.album in library.albums
            ), f"Album {self._configs.album} not found in library {library_name}"
            photos: PhotoAlbum = library.albums[self._configs.album]
            error_handler = build_photos_exception_handler(logger, self.icloud)
            photos.exception_handler = error_handler
            photos_count: int | None = len(photos)

            photos_count, photos_iterator = handle_recent_until_found(
                photos_count, photos, self._configs.recent, self._configs.until_found
            )
            log_at_download_start(
                logger, photos_count, self._configs.size, self._configs.skip_videos, directory
            )
            consecutive_files_found = Counter(0)
            photos_counter = 0
            while True:
                try:
                    if self._status == PolicyStatus.STOPPED:  # policy is interrupted
                        logger.info(f"Policy: {self._name} is interrupted by user. Exiting.")
                        break
                    if should_break(consecutive_files_found, self._configs.until_found):
                        logger.info(
                            "Found %s consecutive previously downloaded photos. Exiting",
                            self._configs.until_found,
                        )
                        break
                    item = next(photos_iterator)
                    download_result = await async_download_photo(consecutive_files_found, item)
                    if download_result and self._configs.delete_after_download:
                        delete_local = partial(
                            delete_photo,
                            logger,
                            self.icloud.photos,
                            library,
                            item,
                        )

                        retrier(delete_local, error_handler)

                    photos_counter += 1
                    if (progress := int(photos_counter / photos_count * 100)) != self._progress:
                        self._progress = progress
                except StopIteration:
                    break

            if self._configs.auto_delete:
                autodelete_photos(
                    logger=logger,
                    dry_run=False,
                    library_object=library,
                    folder_structure=self._configs.folder_structure,
                    directory=directory,
                    primary_sizes=self._configs.size,
                )
        except Exception as e:
            logger.error(f"Error running policy: {self._name}. Exiting.")
            self._status = PolicyStatus.ERRORED
            self._progress = 0
            raise e

        logger.info(
            f"Total of {photos_counter} items in {self._configs.library} from album {self._configs.album} have been downloaded at {self._configs.directory}"
        )
        self._status = PolicyStatus.STOPPED

    def interrupt(self):
        assert self._status == PolicyStatus.RUNNING, "Can only interrupt when policy is running"
        self._status = PolicyStatus.STOPPED



"""Handles file downloads with retries and error handling"""

import datetime
import logging
import os
import socket
import time

from requests import Response
from requests.exceptions import ConnectionError
from tzlocal import get_localzone

# Import the constants object so that we can mock WAIT_SECONDS in tests
from icloudpd import constants
from pyicloud_ipd.asset_version import AssetVersion
from pyicloud_ipd.base import PyiCloudService
from pyicloud_ipd.exceptions import PyiCloudAPIResponseException
from pyicloud_ipd.services.photos import PhotoAsset
from pyicloud_ipd.version_size import VersionSize


def update_mtime(created: datetime.datetime, download_path: str) -> None:
    """Set the modification time of the downloaded file to the photo creation date"""
    if created:
        created_date = None
        try:
            created_date = created.astimezone(get_localzone())
        except (ValueError, OSError):
            # We already show the timezone conversion error in base.py,
            # when generating the download directory.
            # So just return silently without touching the mtime.
            return
        set_utime(download_path, created_date)


def set_utime(download_path: str, created_date: datetime.datetime) -> None:
    """Set date & time of the file"""
    ctime = time.mktime(created_date.timetuple())
    os.utime(download_path, (ctime, ctime))


def mkdirs_for_path(logger: logging.Logger, download_path: str) -> bool:
    """Creates hierarchy of folders for file path if it needed"""
    try:
        # get back the directory for the file to be downloaded and create it if
        # not there already
        download_dir = os.path.dirname(download_path)
        os.makedirs(name=download_dir, exist_ok=True)
        return True
    except OSError:
        logger.error(
            "Could not create folder %s",
            download_dir,
        )
        return False


def mkdirs_for_path_dry_run(logger: logging.Logger, download_path: str) -> bool:
    """DRY Run for Creating hierarchy of folders for file path"""
    download_dir = os.path.dirname(download_path)
    if not os.path.exists(download_dir):
        logger.debug(
            "[DRY RUN] Would create folder hierarchy %s",
            download_dir,
        )
    return True


def download_response_to_path(
    _logger: logging.Logger, response: Response, download_path: str, created_date: datetime.datetime
) -> bool:
    """Saves response content into file with desired created date"""
    temp_download_path = download_path + ".part"
    with open(temp_download_path, "wb") as file_obj:
        for chunk in response.iter_content(chunk_size=1024):
            if chunk:
                file_obj.write(chunk)
    os.rename(temp_download_path, download_path)
    update_mtime(created_date, download_path)
    return True


def download_response_to_path_dry_run(
    logger: logging.Logger,
    _response: Response,
    download_path: str,
    _created_date: datetime.datetime,
) -> bool:
    """Pretends to save response content into a file with desired created date"""
    logger.info(
        "[DRY RUN] Would download %s",
        download_path,
    )
    return True


def download_media(
    logger: logging.Logger,
    dry_run: bool,
    icloud: PyiCloudService,
    photo: PhotoAsset,
    download_path: str,
    version: AssetVersion,
    size: VersionSize,
) -> bool:
    """Download the photo to path, with retries and error handling"""

    mkdirs_local = mkdirs_for_path_dry_run if dry_run else mkdirs_for_path
    download_local = download_response_to_path_dry_run if dry_run else download_response_to_path

    if not mkdirs_local(logger, download_path):
        return False

    for retries in range(constants.MAX_RETRIES):
        try:
            photo_response = photo.download(version.url)
            if photo_response:
                return download_local(logger, photo_response, download_path, photo.created)

            logger.error(
                "Could not find URL to download %s for size %s", version.filename, size.value
            )
            break

        except (ConnectionError, socket.timeout, PyiCloudAPIResponseException) as ex:
            if "Invalid global session" in str(ex):
                logger.error("Session error, re-authenticating...")
                if retries > 0:
                    # If the first re-authentication attempt failed,
                    # start waiting a few seconds before retrying in case
                    # there are some issues with the Apple servers
                    time.sleep(constants.WAIT_SECONDS)

                icloud.authenticate()
            else:
                # you end up here when p.e. throttling by Apple happens
                wait_time = (retries + 1) * constants.WAIT_SECONDS
                logger.error(
                    "Error downloading %s, retrying after %s seconds...", photo.filename, wait_time
                )
                time.sleep(wait_time)

        except OSError:
            logger.error(
                "IOError while writing file to %s. "
                + "You might have run out of disk space, or the file "
                + "might be too large for your OS. "
                + "Skipping this file...",
                download_path,
            )
            break
    else:
        logger.error(
            "Could not download %s. Please try again later.",
            photo.filename,
        )

    return False



from enum import Enum


class RawTreatmentPolicy(Enum):
    AS_IS = "as-is"
    AS_ORIGINAL = "as-original"
    AS_ALTERNATIVE = "as-alternative"

    def __str__(self) -> str:
        return self.name




from typing import Callable, Optional, TypeVar

_Tin = TypeVar("_Tin")
_Tin2 = TypeVar("_Tin2")
_Tin3 = TypeVar("_Tin3")
_Tout = TypeVar("_Tout")


def bind(
    func: Callable[[_Tin], Optional[_Tout]],
) -> Callable[[Optional[_Tin]], Optional[_Tout]]:
    """
    Monadic bind for Optional.

    Example usage:
        >>> def div8(divider: int) -> Optional[float]:
        ...     if divider == 0:
        ...         return None
        ...     return 8 / divider
        >>> b = bind(div8)

        Basic operation on div8:
        >>> div8(2) == 4
        True
        >>> div8(0) == None
        True

        Passing None throws:
        >>> div8(None)
        Traceback (most recent call last):
            ...
        TypeError: unsupported operand type(s) for /: 'int' and 'NoneType'

        Binding works for basic and None:
        >>> b(2) == 4
        True
        >>> b(0) == None
        True
        >>> b(None) == None
        True

    """

    def _intern(input: Optional[_Tin]) -> Optional[_Tout]:
        if input:
            return func(input)
        return None

    return _intern


def lift2(
    func: Callable[[_Tin, _Tin2], _Tout],
) -> Callable[[Optional[_Tin], Optional[_Tin2]], Optional[_Tout]]:
    """
    Lifts regular function into Optional. (Lift2 for Optional Applicative Functor)
    (a -> b -> c) -> Maybe a -> Maybe b -> Maybe c

    Example usage:
        >>> def _mul(input: int, input2: int) -> int:
        ...     return input * input2
        >>> l = lift2(_mul)

        Works for numbers:
        >>> l(2, 3) == 6
        True

        Works for None in any and/or all parameter:
        >>> l(2, None) == None
        True
        >>> l(None, 3) == None
        True
        >>> l(None, None) == None
        True

    """

    def _intern(input: Optional[_Tin], input2: Optional[_Tin2]) -> Optional[_Tout]:
        if input and input2:
            return func(input, input2)
        return None

    return _intern


def lift3(
    func: Callable[[_Tin, _Tin2, _Tin3], _Tout],
) -> Callable[[Optional[_Tin], Optional[_Tin2], Optional[_Tin3]], Optional[_Tout]]:
    """
    Lifts regular function into Optional. see lift2 for Optional Applicative Functor
    (a -> b -> c) -> Maybe a -> Maybe b -> Maybe c
    """

    def _intern(
        input: Optional[_Tin], input2: Optional[_Tin2], input3: Optional[_Tin3]
    ) -> Optional[_Tout]:
        if input and input2 and input3:
            return func(input, input2, input3)
        return None

    return _intern

__all__ = [
    'bind',
    'lift2',
    'lift3',
]


from icloudpd_web.app import create_app
import socketio

# Create the app without static file serving
app, sio = create_app(serve_static=False)

# ASGI app by wrapping FastAPI with Socket.IO
socket_app = socketio.ASGIApp(sio, app)



# mypy: ignore-errors
from __future__ import absolute_import
import os
import uuid
from datetime import datetime
from calendar import monthrange


class ContactsService(object):
    """
    The 'Contacts' iCloud service, connects to iCloud and returns contacts.
    """
    def __init__(self, service_root, session, params):
        self.session = session
        self.params = params
        self._service_root = service_root
        self._contacts_endpoint = '%s/co' % self._service_root
        self._contacts_refresh_url = '%s/startup' % self._contacts_endpoint
        self._contacts_changeset_url = '%s/changeset' % self._contacts_endpoint

    def refresh_client(self, from_dt=None, to_dt=None):
        """
        Refreshes the ContactsService endpoint, ensuring that the
        contacts data is up-to-date.
        """
        params_contacts = dict(self.params)
        params_contacts.update({
            'clientVersion': '2.1',
            'locale': 'en_US',
            'order': 'last,first',
        })
        req = self.session.get(
            self._contacts_refresh_url,
            params=params_contacts
        )
        self.response = req.json()
        params_refresh = dict(self.params)
        params_refresh.update({
            'prefToken': req.json()["prefToken"],
            'syncToken': req.json()["syncToken"],
        })
        self.session.post(self._contacts_changeset_url, params=params_refresh)
        req = self.session.get(
            self._contacts_refresh_url,
            params=params_contacts
        )
        self.response = req.json()

    def all(self):
        """
        Retrieves all contacts.
        """
        self.refresh_client()
        return self.response['contacts']



# mypy: ignore-errors
from datetime import datetime
import sys


class UbiquityService(object):
    """ The 'Ubiquity' iCloud service."""

    def __init__(self, service_root, session, params):
        self.session = session
        self.params = params
        self._root = None

        self._service_root = service_root
        self._node_url = '/ws/%s/%s/%s'

    def get_node_url(self, id, variant='item'):
        return self._service_root + self._node_url % (
            self.params['dsid'],
            variant,
            id
        )

    def get_node(self, id):
        request = self.session.get(self.get_node_url(id))
        return UbiquityNode(self, request.json())

    def get_children(self, id):
        request = self.session.get(
            self.get_node_url(id, 'parent')
        )
        items = request.json()['item_list']
        return [UbiquityNode(self, item) for item in items]

    def get_file(self, id, **kwargs):
        request = self.session.get(
            self.get_node_url(id, 'file'),
            **kwargs
        )
        return request

    @property
    def root(self):
        if not self._root:
            self._root = self.get_node(0)
        return self._root

    def __getattr__(self, attr):
        return getattr(self.root, attr)

    def __getitem__(self, key):
        return self.root[key]


class UbiquityNode(object):
    def __init__(self, conn, data):
        self.data = data
        self.connection = conn

    @property
    def item_id(self):
        return self.data.get('item_id')

    @property
    def name(self):
        return self.data.get('name')

    @property
    def type(self):
        return self.data.get('type')

    def get_children(self):
        if not hasattr(self, '_children'):
            self._children = self.connection.get_children(self.item_id)
        return self._children

    @property
    def size(self):
        try:
            return int(self.data.get('size'))
        except ValueError:
            return None

    @property
    def modified(self):
        return datetime.strptime(
            self.data.get('modified'),
            '%Y-%m-%dT%H:%M:%SZ'
        )

    def dir(self):
        return [child.name for child in self.get_children()]

    def open(self, **kwargs):
        return self.connection.get_file(self.item_id, **kwargs)

    def get(self, name):
        return [
            child for child in self.get_children() if child.name == name
        ][0]

    def __getitem__(self, key):
        try:
            return self.get(key)
        except IndexError:
            raise KeyError('No child named %s exists' % key)

    def __unicode__(self):
        return self.name

    def __str__(self):
        as_unicode = self.__unicode__()
        if sys.version_info[0] >= 3:
            return as_unicode
        else:
            return as_unicode.encode('ascii', 'ignore')

    def __repr__(self):
        return "<%s: '%s'>" % (
            self.type.capitalize(),
            self
        )
